"""Generate a dataset of expressions and their unique like terms as string outputs.

The challenge then is for the model to identify and extract the like terms from an
expression, e.g.

IN: 4x + x^3 - 7x
OUT: 4x[SEP]-7x
"""
import os
import random
from pathlib import Path
from typing import Dict, List, Optional, Set
from mathy.expressions import SubtractExpression
from collections import OrderedDict
from mathy.tree import LEFT, RIGHT
import numpy as np
import typer
from tqdm.auto import tqdm

from mathy import ExpressionParser, MathExpression, TermEx, get_term_ex, get_terms
from mathy.problems import gen_simplify_multiple_terms, mathy_term_string

parser = ExpressionParser()


def generate_newline_q_a(
    file_base: str,
    number: int,
    exclude: Optional[Set[str]] = None,
    eval: bool = False,
    max_len: int = 128,
):

    train_file = f"{file_base}.txt"
    if exclude is None:
        exclude = set()
    skips = 0
    skip_threshold = 100
    problems = 0
    min_like = 4 if eval else 2
    max_like = 12 if eval else 4
    min_noise = 1
    max_noise = 2
    with Path(train_file).open("w") as f:
        with tqdm(total=number, mininterval=0.25, desc=file_base) as pbar:
            while problems < number:
                text, complexity = gen_simplify_multiple_terms(
                    random.randint(min_like, max_like),
                    noise_probability=0.5,
                    noise_terms=random.randint(min_noise, max_noise),
                    op=["+", "-"],
                )

                if text in exclude or len(text) >= max_len:
                    skips += 1
                    if skips >= skip_threshold:
                        raise ValueError(
                            f"Failed to generate more unique problems after {skips} tries!"
                        )
                    continue

                skips = 0
                exclude.add(text)
                answer = list_like_terms(text, sep=",")
                f.write(f"{text}\n{answer}\n")
                pbar.update(1)
                problems += 1


def list_like_terms(input_problem: str, sep: str) -> str:
    expression: MathExpression = parser.parse(input_problem)
    term_nodes: List[MathExpression] = get_terms(expression)
    node_groups: OrderedDict[str, List[MathExpression]] = OrderedDict()
    for term_node in term_nodes:
        ex: Optional[TermEx] = get_term_ex(term_node)
        assert ex is not None, f"invalid expression {term_node}"
        key = mathy_term_string(variable=ex.variable, exponent=ex.exponent)
        if key == "":
            key = "const"
        if key not in node_groups:
            node_groups[key] = [term_node]
        else:
            node_groups[key].append(term_node)

    # TODO: Sort terms by exponent then variables to produce a "natural" ordering and combine keys
    out_terms: List[str] = []
    for term_node in term_nodes:
        ex: Optional[TermEx] = get_term_ex(term_node)
        assert ex is not None, f"invalid expression {term_node}"
        key = mathy_term_string(variable=ex.variable, exponent=ex.exponent)
        if key == "":
            key = "const"
        assert key in node_groups
        if len(node_groups[key]) <= 1:
            continue
        is_parent_sub = isinstance(term_node.parent, SubtractExpression)
        is_right_side = term_node.parent.get_side(term_node) == RIGHT
        pre = "-" if is_parent_sub and is_right_side else ""
        out_terms.append(f"{pre}{term_node}")
    return sep.join(out_terms)


def main(
    name: str,
    train_size: int = 1000 * 1000,
    eval_size: int = 1000,
    max_len: int = 128,
    include_eval: bool = True,
    include_generalization: bool = True,
):
    current = os.path.dirname(__file__)
    train_file = os.path.join(current, f"{name}.train")
    eval_file = os.path.join(current, f"{name}.eval")
    generalization_file = os.path.join(current, f"{name}.generalization")

    seen_texts: Set[str] = set()

    print(f"Generating train dataset with {train_size} examples...")
    generate_newline_q_a(train_file, train_size, seen_texts, max_len=max_len)
    if include_eval:
        print(f"Generating eval dataset with {eval_size} examples...")
        generate_newline_q_a(eval_file, eval_size, seen_texts, max_len=max_len)

    if include_generalization:
        print(f"Generating generalization dataset with {eval_size} examples...")
        generate_newline_q_a(
            generalization_file, eval_size, seen_texts, max_len=max_len, eval=True,
        )


if __name__ == "__main__":
    typer.run(main)



import sys

# Use typing_extensions for Python < 3.8
if sys.version_info < (3, 8):
    from typing_extensions import Final, Literal  # noqa
else:
    from typing_extensions import Final, Literal  # noqa




"""Utility for generating question/answer datasets of two terms that are either like or unlike"""
import os
import random
from pathlib import Path
from typing import Dict, List, Optional, Set, Tuple

import numpy as np
import typer
from tqdm.auto import tqdm

from mathy import ExpressionParser, MathExpression, TermEx, get_term_ex, terms_are_like
from mathy.problems import (
    MathyTermTemplate,
    gen_simplify_multiple_terms,
    get_rand_term_templates,
    mathy_term_string,
    rand_bool,
    rand_op,
)

parser = ExpressionParser()


def generate_newline_q_a(
    file_base: str,
    number: int,
    exclude: Optional[Set[str]] = None,
    eval: bool = False,
    max_len: int = 128,
):

    train_file = f"{file_base}.txt"
    if exclude is None:
        exclude = set()
    skips = 0
    skip_threshold = 50000
    problems = 0
    half = number // 2
    positives: int = 0
    negatives: int = 0
    with Path(train_file).open("w") as f:
        with tqdm(total=number, mininterval=0.25, desc=file_base) as pbar:
            while problems < number:
                text, answer = make_problem(positives >= half, negatives >= half)
                if text in exclude or len(text) >= max_len:
                    skips += 1
                    if skips >= skip_threshold:
                        raise ValueError(
                            f"Failed to generate more unique problems after {skips} tries!"
                        )
                    continue
                # Make sure the dataset is split 50/50 for labels
                if answer == 2:
                    if positives >= half:
                        continue
                    positives += 1
                elif answer == 0:
                    if negatives >= half:
                        continue
                    negatives += 1

                skips = 0
                exclude.add(text)
                f.write(f"{text}\n{answer}\n")
                pbar.update(1)
                problems += 1


def make_problem(
    skip_positives: bool = False, skip_negatives: bool = False
) -> Tuple[str, int]:
    left_tpl, right_tpl = get_rand_term_templates(2, exponent_probability=0.5)
    left_base = f"{left_tpl.variable}^{left_tpl.exponent}"
    assert left_base != right_tpl
    problem_type = random.randint(0, 3)
    if problem_type < 2 and skip_negatives:
        problem_type = 2
    elif problem_type == 2 and skip_positives:
        problem_type = 0

    if problem_type == 0:
        # Generate a problem with 0 like terms, and a random operator
        operator = random.choice(["+", "-", "*"])
        problem = f"{left_tpl.make()} {operator} {right_tpl.make()}"
        count = 0
    elif problem_type == 1:
        # Generate a problem with like variables but unlike exponents
        operator = random.choice(["+", "-", "*"])
        other_exp = None if left_tpl.exponent is not None else random.randint(1, 12)
        right_tpl = MathyTermTemplate(variable=left_tpl.variable, exponent=other_exp)
        problem = f"{left_tpl.make()} {operator} {right_tpl.make()}"
        count = 0
    elif problem_type == 2 or problem_type == 3:
        # Generate a problem with 2 like terms, and a random operator
        operator = random.choice(["+", "-"])
        problem = f"{left_tpl.make()} {operator} {left_tpl.make()}"
        count = 2
    else:
        raise ValueError(f"Unknown problem type: {problem_type}")
    return problem, count


def main(
    name: str,
    train_size: int = 100 * 1000,
    eval_size: int = 1000,
    max_len: int = 32,
    include_eval: bool = True,
    include_generalization: bool = True,
):
    current = os.path.dirname(__file__)
    train_file = os.path.join(current, f"{name}.train")
    eval_file = os.path.join(current, f"{name}.eval")
    generalization_file = os.path.join(current, f"{name}.generalization")

    seen_texts: Set[str] = set()

    print(f"Generating train dataset with {train_size} examples...")
    generate_newline_q_a(train_file, train_size, seen_texts, max_len=max_len)
    if include_eval:
        print(f"Generating eval dataset with {eval_size} examples...")
        generate_newline_q_a(eval_file, eval_size, seen_texts, max_len=max_len)

    if include_generalization:
        print(f"Generating generalization dataset with {eval_size} examples...")
        generate_newline_q_a(
            generalization_file, eval_size, seen_texts, max_len=max_len, eval=True,
        )


if __name__ == "__main__":
    typer.run(main)



import os
import random
from pathlib import Path
from typing import Dict, List, Optional, Set
import numpy as np
import typer
from tqdm.auto import tqdm

from mathy import ExpressionParser, MathExpression, TermEx, get_term_ex, get_terms
from mathy.problems import gen_simplify_multiple_terms, mathy_term_string, use_pretty_numbers

parser = ExpressionParser()


def generate_newline_q_a(
    file_base: str,
    number: int,
    exclude: Optional[Set[str]] = None,
    eval: bool = False,
    max_len: int = 128,
):

    train_file = f"{file_base}.txt"
    if exclude is None:
        exclude = set()
    skips = 0
    skip_threshold = 100
    problems = 0
    min_like = 4 if eval else 2
    max_like = 12 if eval else 8
    with Path(train_file).open("w") as f:
        with tqdm(total=number, mininterval=0.25, desc=file_base) as pbar:
            while problems < number:
                text, complexity = gen_simplify_multiple_terms(
                    random.randint(min_like, max_like),
                    noise_probability=0.5,
                    noise_terms=random.randint(1, 5),
                    op=["+", "-"],
                )

                if text in exclude or len(text) >= max_len:
                    skips += 1
                    if skips >= skip_threshold:
                        raise ValueError(
                            f"Failed to generate more unique problems after {skips} tries!"
                        )
                    continue

                skips = 0
                exclude.add(text)
                answer = count_like_terms(text)
                f.write(f"{text}\n{answer}\n")
                pbar.update(1)
                problems += 1


def count_like_terms(input_problem: str) -> int:
    expression: MathExpression = parser.parse(input_problem)
    term_nodes: List[MathExpression] = get_terms(expression)
    node_groups: Dict[str, List[MathExpression]] = {}
    for term_node in term_nodes:
        ex: Optional[TermEx] = get_term_ex(term_node)
        assert ex is not None, f"invalid expression {term_node}"
        key = mathy_term_string(variable=ex.variable, exponent=ex.exponent)
        if key == "":
            key = "const"
        if key not in node_groups:
            node_groups[key] = [term_node]
        else:
            node_groups[key].append(term_node)
    like_terms = 0
    for k, v in node_groups.items():
        if len(v) <= 1:
            continue
        like_terms += len(v)
    return like_terms


def main(
    name: str,
    train_size: int = 200 * 1000,
    eval_size: int = 1000,
    max_len: int = 128,
    include_eval: bool = True,
    include_generalization: bool = True,
    pretty:bool = False
):

    use_pretty_numbers(pretty)

    current = os.path.dirname(__file__)
    train_file = os.path.join(current, f"{name}.train")
    eval_file = os.path.join(current, f"{name}.eval")
    generalization_file = os.path.join(current, f"{name}.generalization")

    seen_texts: Set[str] = set()

    print(f"Generating train dataset with {train_size} examples...")
    generate_newline_q_a(train_file, train_size, seen_texts, max_len=max_len)
    if include_eval:
        print(f"Generating eval dataset with {eval_size} examples...")
        generate_newline_q_a(eval_file, eval_size, seen_texts, max_len=max_len)

    if include_generalization:
        print(f"Generating generalization dataset with {eval_size} examples...")
        generate_newline_q_a(
            generalization_file, eval_size, seen_texts, max_len=max_len, eval=True,
        )


if __name__ == "__main__":
    typer.run(main)



import os
import random
from pathlib import Path
from typing import Dict, List, Optional, Set
import numpy as np
import typer
from tqdm.auto import tqdm
from mathy.problems import rand_number


def generate_problems(
    file_base: str, number: int, exclude: Optional[Set[str]] = None, eval: bool = False
):
    inputs_file_name = f"{file_base}.inputs.txt"
    outputs_file_name = f"{file_base}.outputs.txt"
    if exclude is None:
        exclude = set()
    skips = 0
    skip_threshold = 1200
    problems = 0

    with Path(outputs_file_name).open("w") as outputs_file:
        with Path(inputs_file_name).open("w") as inputs_file:
            with tqdm(total=number, mininterval=0.25, desc=file_base) as pbar:
                while problems < number:
                    one = rand_number()
                    two = rand_number()
                    text = f"{one}+{two}"
                    answer = f"{one + two}"
                    if text in exclude:
                        skips += 1
                        if skips >= skip_threshold:
                            raise ValueError(
                                f"Failed to generate more unique problems after {skips} tries!"
                            )
                            continue
                    exclude.add(text)
                    outputs_file.write(f"{answer}\n")
                    inputs_file.write(f"{text}\n")
                    pbar.update(1)
                    problems += 1


def main(
    name: str,
    train_size: int = 100 * 1000,
    eval_size: int = 10 * 1000,
    include_eval: bool = True,
    include_generalization: bool = True,
):
    current = os.path.dirname(__file__)
    train_file = os.path.join(current, f"{name}.train")
    eval_file = os.path.join(current, f"{name}.eval")
    generalization_file = os.path.join(current, f"{name}.generalization")

    seen_texts: Set[str] = set()

    print(f"Generating train dataset with {train_size} examples...")
    generate_problems(train_file, train_size, seen_texts, eval=True)
    if include_eval:
        print(f"Generating eval dataset with {eval_size} examples...")
        generate_problems(eval_file, eval_size, seen_texts, eval=True)

    if include_generalization:
        print(f"Generating generalization dataset with {eval_size} examples...")
        generate_problems(generalization_file, eval_size, seen_texts)


if __name__ == "__main__":
    typer.run(main)
